##This is file that contains collection of functions for asg simulation study.
##Created by H.Chun Sept 14. 2013
###############################################################################

library(mvtnorm)
library(kernlab)
library(corpcor)
library(glasso)


##1. function for creating kernels

createkernel <- function(y=y, kernel.sigma= 1e-2){

  rbf <- rbfdot(sigma = kernel.sigma)
  n <- dim(y)[1]
  p <- dim(y)[2]
  karray <- array(NA, dim=c(p, n, n))
  for(i in 1:p){
    karray[i,,] <- kernelMatrix(rbf, x=as.vector(y[,i]))
  }
  return(karray)
}



##2. expand array

expandarray <- function(karray2, i, j){
	if(i==0 & j==0){
		subarray <- karray2
		adim <- dim(subarray)[1]
		bdim <- dim(subarray)[2]
		if(length(dim(subarray)) >= 3){
			expandarray2 <- matrix(NA, nr = bdim, nc= (adim*bdim))
			for(l in 1:adim){
				expandarray2[,((l-1)*bdim+1):(l*bdim)]<-subarray[l,,]
			}
		}else{
			expandarray2 <- subarray
		}
	}else{
		subarray <- karray2[-c(i,j),,]

		adim <- dim(subarray)[1]
		bdim <- dim(subarray)[2]

		if(length(dim(subarray)) >= 3){
			expandarray2 <- matrix(NA, nr = bdim, nc= (adim*bdim))
			for(l in 1:adim){
				expandarray2[,((l-1)*bdim+1):(l*bdim)]<-subarray[l,,]
			}
		}else{
			expandarray2 <- subarray
		}
	}
  return(t(expandarray2))
}




##3. main function for ASG

##Use two tuning parameter epsilon1 and epsilon2
fit.asg <- function(z=z,epsilon1 =5e-2, epsilon2= 5e-2, LINPACK=FALSE){
##step1
##normalize
  z <- scale(z, center=TRUE, scale=TRUE)
  kernel.sigma <- 1/dim(z)[2]/9
##step2
## construct kernel
  karray2 <- createkernel(y=z, kernel.sigma=kernel.sigma)

##step 3 compute the precision matrix
  n <- dim(z)[1]
  Q <- diag(rep(1,n))  - matrix(1, nc=n, nr=n)/n
  p <- dim(z)[2]
  asg <- matrix(1, nc=p, nr=p)
	
  for(i in 1:(p-1)){
    K1 <- karray2[i,,]
    A<- zapsmall(K1%*%Q%*%K1 + epsilon1*diag(rep(1,n)),9)
    if (any(abs(A - t(A)) > 100 * .Machine$double.eps)) A <- (A+t(A))/2 
    inv0.5 <- mpower(A, -0.5)
    for(j in (i+1):p){
      if(i != j){
        K2 <- karray2[j,,]
        K3 <- expandarray(karray2, i, j)
        B <-zapsmall(K2%*%Q%*%K2 + epsilon1*diag(rep(1,n)),9)
        if (any(abs(B - t(B)) > 100 * .Machine$double.eps)) B <- (B+t(B))/2 
        inv20.5<- mpower(B, -0.5)
        svdK3 <- svd(K3%*%Q, LINPACK = TRUE)        
        R <- svdK3$v[,1:(n-1)]
        D <- svdK3$d[1:(n-1)]
        halfdelta <- inv0.5%*%K1%*%(Q-R%*%diag(D^2/(D^2+epsilon2))%*%t(R))%*%K2%*%inv20.5
        asg[i,j] <- asg[j,i]<-svd(halfdelta, LINPACK=LINPACK)$d[1]
      }
    }
  }
  return(list(asg=asg))  
}



###function for ASG2

##4. main function for ASG
fit.asg2 <- function(z=z,epsilon1 =5e-2, epsilon2= 5e-2, LINPACK=FALSE){
##step1
##normalize
  z <- scale(z, center=TRUE, scale=TRUE)
  kernel.sigma <- 1/dim(z)[2]/9
##step2
## construct kernel
  karray2 <- createkernel(y=z, kernel.sigma=kernel.sigma)
  K <- expandarray(karray2, 0,0)
  n <- dim(z)[1]
  Q <- diag(rep(1,n))  - matrix(1, nc=n, nr=n)/n

  mat <- K%*%Q
	
  mateig <- svd(mat, LINPACK=LINPACK)
  R1 <- mateig$v[,1:(n-1)]
  D <- diag((mateig$d[1:(n-1)]))
  d <- (mateig$d[1:(n-1)])
	
##step 3 compute the precision matrix

  p <- dim(z)[2]
  asg <- matrix(100, nc=p, nr=p)
  for(i in 1:(p-1)){
    K1 <- karray2[i,,]
##symmetrization
    A<- zapsmall(K1%*%Q%*%K1 + epsilon1*diag(rep(1,n)),9)
    if (any(abs(A - t(A)) > 100 * .Machine$double.eps)) A <- (A+t(A))/2 
    A0.5 <- mpower(A, 0.5)
    for(j in (i+1):p){
      if(i != j){
        K2 <- karray2[j,,]
        mat2 <- K2%*%Q%*%R1%*%diag(1/(d^2*(d^2+epsilon2))- 1/d^2/epsilon2)%*%t(R1)%*%Q%*%K1
        B<- zapsmall(K2%*%Q%*%K2 + epsilon1*diag(rep(1,n)),9)
        if (any(abs(B - t(B)) > 100 * .Machine$double.eps)) B <- (B+t(B))/2 
        B0.5 <- mpower(B,0.5)
        delta <- B0.5%*%mat2%*%A0.5
        svdres <- svd(delta)$d[1]
        asg[i,j] <- asg[j,i]<-svdres
      }
    }
  }
  return(list(asg=asg))  
}


gcv1 <- function(z=z, evec=5*10^c(-4:3), norm="matrixL2", LINPACK=FALSE){
 
  z <- scale(z, center=TRUE, scale=TRUE)
  kernel.sigma <- 1/dim(z)[2]/9
 
  n <- dim(z)[1]
  p <- dim(z)[2]
 
  karray2 <- createkernel(y=z, kernel.sigma=kernel.sigma)
  
  tot <- matrix(NA, nc=length(evec), nr= p)

  if(norm=="frob"){
    for(i in 1:dim(z)[2]){
      for(k in 1:length(evec)){
        epsilon1 <- evec[k]
        Li <- rbind(rep(1, n),karray2[i,,])
        Lci   <- c()
        for(j in 1:p){
          if(i != j){
            Lci <- rbind(Lci, karray2[j,,])
          }
        }
        Lci <- rbind(rep(1, dim(Lci)[2]), Lci)
        
        svres <- svd(Li, LINPACK=LINPACK)
        proj <- svres$v%*%diag(svres$d^2/(svres$d^2+epsilon1))%*%t(svres$v)
        trs <- sum(svres$d^2/(svres$d^2+epsilon1))/dim(proj)[1]
        tot[i,k] <- sum(((Lci) - (Lci)%*%proj)^2)/(1-trs)^2
      }
    }
  }else{
    for(i in 1:dim(z)[2]){
      for(k in 1:length(evec)){
        epsilon1 <- evec[k]
        Li <- rbind(rep(1, n),karray2[i,,])
        Lci   <- c()
        for(j in 1:p){
          if(i != j){
            Lci <- rbind(Lci, karray2[j,,])
          }
        }
        Lci <- rbind(rep(1, dim(Lci)[2]), Lci)
        
        svres <- svd(Li, LINPACK=LINPACK)
        proj <- svres$v%*%diag(svres$d^2/(svres$d^2+epsilon1))%*%t(svres$v)
        trs <- sum(svres$d^2/(svres$d^2+epsilon1))/dim(proj)[1]
        tot[i,k] <- svd((Lci - Lci%*%proj), LINPACK=LINPACK)$d[1]^2/(1-trs)^2
      }
    }
  }
  return(apply(tot, 2, sum))
}





gcv2 <- function(z=z, evec=evec,norm="matrixL2", LINPACK=FALSE){
 
 z <- scale(z, center=TRUE, scale=TRUE)
 kernel.sigma <- 1/dim(z)[2]/9
 

 n <- dim(z)[1]
 p <- dim(z)[2]
  
 karray2 <- createkernel(y=z, kernel.sigma=kernel.sigma)
  
 tot <- matrix(0, nc=length(evec), nr=p)

 if(norm=="frob"){
   for(i in 1:p){
     for(j in 1:p){
       if(i != j){       
         Lij <- rbind(rep(1, n),karray2[i,,], karray2[j,, ])
         Lcij  <- c()
         for(k in (1:p)[-c(i,j)]){
           Lcij <- rbind(Lcij, karray2[k,,])
         }
       
         Lcij <- rbind(rep(1, dim(Lcij)[2]), Lcij)
         svres <- svd(Lcij, LINPACK=LINPACK)       

       for(k in 1:length(evec)){
         epsilon2 = evec[k]
         proj <- svres$v%*%diag(svres$d^2/(svres$d^2+epsilon2))%*%t(svres$v)
         trs <- sum(svres$d^2/(svres$d^2+epsilon2))/(dim(proj)[1])
         tot[i,k] <- tot[i,k] +sum(((Lij) - (Lij)%*%proj)^2)/(1-trs)^2
       }
       }
     }
   }
 }else{
   for(i in 1:p){
     for(j in 1:p){
       if(i > j){       
         Lij <- rbind(rep(1, n),karray2[i,,], karray2[j,, ])
         Lcij  <- c()
         for(k in (1:p)[-c(i,j)]){
           Lcij <- rbind(Lcij, karray2[k,,]) 
         }
       
         Lcij <- rbind(rep(1, dim(Lcij)[2]), Lcij)
         svres <- svd(Lcij, LINPACK=LINPACK)       
      
         for(k in 1:length(evec)){
           epsilon2 = evec[k]
           proj <- svres$v%*%diag(svres$d^2/(svres$d^2+epsilon2))%*%t(svres$v)
           trs <- sum(svres$d^2/(svres$d^2+epsilon2))/(dim(proj)[1])
           tot[i,k] <- tot[i,k] +svd(((Lij) - (Lij)%*%proj))$d[1]^2/(1-trs)^2
         }
       }
     }
   } 
 }
 return(apply(tot,2,sum))
}



##threshold estimator
threshold.est<- function(est.precision = est.precision, vlambda= seq(0, 1, length=vlambda.length)){
est <- list()
for(i in 1:length(vlambda)){
  est[[i]] <- abs(est.precision) >= vlambda[i]
}
return(est)
} 






##function for computing ROC
computeROC <- function(est, vlambda.length=vlambda.length, true=true){
roc <- matrix(NA, nr=vlambda.length, nc=2)
for(i in 1:vlambda.length){
  est1 <- est[[i]] !=0
  diag(est1) <- 1
  tp <- (sum(true*est1, na.rm=T)-p)/(sum(true)-p)
  fp <-  sum( (1-true)*est1, na.rm=T)/(sum(1-true))
  roc[i,] <- c(fp, tp) 
}
return(roc)
}



plotaverage <- function(mat=mat, add=F, col=1, lty=1, ...){
 if(!add){
 plot(apply(mat[,1,], 1, mean), apply(mat[,2,], 1, mean), xlab="1-specificity", ylab="sensitivity", col=col, type= "l", lty=lty,... )
}else{
 points(apply(mat[,1,], 1, mean), apply(mat[,2,], 1, mean), col=col, type="l",lty=lty,...)
}
}







##7. Function for copula ggm with Spearman's rho

fit.rho <- function(z=z, lambda=lambda, adaptive =TRUE, bic=F){
  rho.mat <- (2*sin(pi/6*cor(z, method="spearman")))

  eival <- eigen(rho.mat)$val[dim(rho.mat)[1]]
  if(eival < 0 ) rho.mat <- rho.mat -diag(rep(eival-1e-6, dim(rho.mat)[1]))

  
  weight <- solve(rho.mat)
  n <- dim(z)[1]
  p <- dim(z)[2]
  if(!bic){
   if(adaptive){
     fit <- glasso(rho.mat, lambda/(1e-32+abs(weight)), penalize.diagonal=TRUE)

   }else{
     fit <- glasso(rho.mat, lambda,  penalize.diagonal=TRUE)
   }
   return(fit$wi)
 }else{

   if(adaptive){
     fit <- glasso(rho.mat, lambda/(1e-32+abs(weight)), penalize.diagonal=TRUE)
     
   }else{
     fit <- glasso(rho.mat, lambda,  penalize.diagonal=TRUE)
   }
   bic =  -log(det(fit$wi)) + sum(rho.mat*fit$wi) +  (log(n) + log(p))/n*((sum(zapsmall(fit$wi,6)!=0)-p)/2+p) 
   return(list(fit=fit, bic=bic))
   
 }
}





fit.regular <- function(z=z, lambda=lambda, adaptive =TRUE, bic=F){
  cor <-cor(z)


  eival <- eigen(cor)$val[dim(cor)[1]]
  if(eival < 0 ) cor <- cor -diag(rep(eival-1e-6, dim(cor)[1]))
  





  
  weight <- solve(cor(z))
  
  n <- dim(z)[1]
  p <- dim(z)[2]
  
  if(!bic){
    
    if(adaptive){
      fit <- glasso(cor, lambda/(1e-32 + abs(weight)), penalize.diagonal=TRUE)
      
    }else{
      fit <- glasso(cor, lambda,penalize.diagonal=TRUE)
  }
    return(fit$wi)
  }else{
    
    if(adaptive){
    fit <- glasso(cor, lambda/(1e-32 + abs(weight)), penalize.diagonal=TRUE)
    
  }else{
    fit <- glasso(cor, lambda,penalize.diagonal=TRUE)
  }
    bic =  -log(det(fit$wi)) + sum(cor*fit$wi) +  (log(n) + log(p))/n*((sum(zapsmall(fit$wi,6)!=0)-p)/2+p) 
    return(list(fit=fit, bic=bic))
  }
}




computeAUC <- function(resultmat){
  resultmat <- rbind(c(1,0),resultmat) 
  m <- dim(resultmat)[1]
  
  power <- 1 - resultmat[,2]
  mpower <-  (power[-1] + power[-m])/2
  xx <- resultmat[,1]
  d <- -xx[-1] + xx[-m] 
  area <- sum(mpower*d)
  return(area)
}
